from fireo.fields.errors import FieldNotFound, MissingFieldOptionError
from fireo.managers import managers
from fireo import fields
from fireo.models.errors import NonAbstractModel, UnSupportedMeta
from fireo.utils import utils


class ModelMeta(type):
    """Create fields, manager and other stuff for Models

    Add `_meta` attribute to each Model which hold all the information about fields, managers and any
    other stuff related to model. Also responsible to generate collection name from model.

    Methods
    -------
    __new__(mcs, name, base, attrs):
        Create modified type for model class
    """

    def __new__(mcs, name, base, attrs):
        """Create modified type for model class

        Convert fields, managers and other stuff into `Meta` class and attached it with model class

        Attributes
        ----------
        _meta : Meta()
            Hold all information about you model, can be accessible via class **cls._meta**

        collection_name : str
            Name of collection which is saved in database if user not provided any collection name
            then `Meta` class generate it from Model name and attach it with class can be accessible
            via class **cls.collection_name**

        Returns
        -------
            Modified type for model class
        """
        cls = super().__new__(mcs, name, base, attrs)

        # check user specify any additional meta data for this model or not
        # For example collection name etc if there is not then assign None to it
        if 'Meta' not in attrs:
            cls.Meta = None

        class Meta:
            """Hold information about fields, manager and other model related stuff

            Meta class get fields, manager and other model related stuff and attach it to
            with model class.

            For example:
                class User:
                    name = TextField()
                    age = NumberField()

                u = User()
                u.name = "Azeem"
                u.age = 25

                This Meta class is responsible to attach these fields with model class
                later on these fields can be accessible via **cls._meta**

                For example:
                    cls._meta.field_list() will return {name: <obj of TextField ox13>, age: <obj of NumberField 0x17>}

                Later on you can use these fields to get corresponding value from model and use them further
                field value can be get like this:

                    getattr(self, field_name)

                Here `self` is instance of model class and `field_name` is one the name from `field_list()`

                In above example if you run this

                    getattr(model_instance, name)
                    getattr(model_instance, age)

                it will return **Azeem** and **25**

            Attributes
            ----------
            field_list : dict
                Contain all model field

            id : str, optional
                Model id if not specify then it return `None` and later create by firestore
                and attach to model

            abstract: bool
                Model is abstract or not

            missing_field: str
                Model config what to do with fields that are comming from firestore and not in model
                possible values are (merge, ignore, raise_error) merge is default

            to_lowercase: bool
                Firestore is case sensitive convert all value in lowercase if set True. By default
                it set False.

            collection : manager
                Collection is class level attribute can be used to access the manager

            See Also
            --------
            fireo.managers.Manager :
                Manager are responsible to run certain functions on firestore like creating collection
                getting documents update etc

            Methods
            -------
            add_model_id(field):
                If user specify any id then attach this id to model

            add_field(field):
                All all user specified fields in Model class

            get_field(name):
                Get field from model on the base of name

            get_field_by_column_name(name):
                Get field according to firestore column name(field name)

            set_user_defined_meta(user_meta):
                Set user defined meta in model class, These meta is actually the config for model

            Raises:
            -------
            FieldNotFound:
                if there is no such in field in model class

            NonAbstractModel:
                If any model is inherit from non abstract model
            """
            field_list = {}  # Hold all the model fields
            id = None  # Model id if user specify otherwise just None will generate late by firestore automatically

            def __init__(self):
                # Convert Model class into collection name
                # change it to lower case and snake case
                # e.g UserCollection into user_collection
                self.collection_name = utils.collection_name(cls.__name__)
                self.abstract = False
                self.missing_field = 'merge'
                self.ignore_none_field = True
                self.to_lowercase = False
                self._referenceDoc = None
                self._firestore_create_time = None
                self._firestore_update_time = None

            # Attached manager to model class
            # later on manager can be accessible via class `collection` attribute
            # like this **cls.collection**
            if 'collection' not in cls.__dict__:
                manager = managers.Manager()
                manager.contribute_to_model(cls)

            def add_model_id(self, field):
                """ Attach user define id to model

                If user specify any custom id for this model
                attach it to model this contains two thing custom id and if field

                For Example:
                    class User:
                        user_id = IDField()

                    In this case it will hold (user_id, <obj of IDField 0x137>)

                    later `id` can be accessible via `_meta`
                    like this **cls._meta.id**

                Parameters
                ----------
                field : IDField()
                    User defined custom id field
                """
                self.id = (field.name, field)

            def add_field(self, field):
                """Add model fields into model meta class

                These fields can be used later for getting corresponding values from
                model instance and some other operations

                can be accessible via `_meta`
                like this **cls._meta.field_list()**

                Parameters
                ----------
                field : Field
                    This can be any type which is derived from `Field` sub class
                    For example `TextField`, `NumberField` etc
                """
                self.field_list[field.name] = field

            def get_field(self, name):
                """Get model field from field list

                Get field on the base of it's name can be accessible via `_meta`
                like this **cls._meta.get_field(name)**

                Parameters
                ---------
                name:
                    Field name

                Returns
                ------
                    Model field

                Raises
                ------
                FieldNotFound:
                    If field not found in model
                """
                if name in self.field_list:
                    return self.field_list[name]
                raise FieldNotFound(f'Field "{name}" not found in model "{cls.__name__}"')

            def get_field_by_column_name(self, name):
                """Get field by column name

                User can also define different name for each field to save in firestore
                Check if user specify any different name then get field corresponding to
                this db_column_name. If no column name is specify for field then just
                return the same field name

                Parameters
                ----------
                name : str
                    column name (field name) coming form firestore

                Raises
                -------
                FieldNotFound:
                    if field not found in model class and model config for `missing_field` is **raise_error**
                """
                for field in self.field_list.values():
                    if name in [field.name, field.db_column_name]:
                        return field
                if self.missing_field == 'merge':
                    f = fields.Field()
                    f.name = name
                    return f
                if self.missing_field == 'ignore':
                    return None
                if self.missing_field == 'raise_error':
                    raise FieldNotFound(f'Field "{name}" not found in model "{cls.__name__}"')

            def set_reference_doc(self, referenceDoc):
                """save the firestore reference doc for further processing"""
                self._referenceDoc = referenceDoc

            def set_user_defined_meta(self, user_meta):
                """Set user defined meta attributes for model

                User can define config by using `Meta` class in models

                Examples
                --------
                .. code-block:: python
                    class User(Model):
                        name = TextField()

                        class Meta:
                            collection_name = "user_test_collection"

                    # All the documents now save in *user_test_collection* for this model

                Parameters
                ---------
                user_meta:
                    User defined meta for model class

                Raises
                -------
                MissingFieldOptionError:
                    If option for missing_field is other than ignore, merge or raise_error
                """
                for name, val in user_meta.__dict__.items():
                    supported_meta = ['collection_name', 'abstract', 'to_lowercase', 'missing_field', 'ignore_none_field']

                    # check if name is supported by meta and name is not
                    # any special name for example '__main__, __doc__'
                    if name not in supported_meta and '__' not in name:
                        raise UnSupportedMeta(f'Meta "{name}" is not recognize in model "{cls.__name__}" '
                                              f'Possible value are {", ".join(supported_meta)}')

                    if name == 'collection_name':
                        self.collection_name = val
                    if name == 'abstract':
                        self.abstract = val
                    if name == 'to_lowercase':
                        self.to_lowercase = val
                    if name == 'ignore_none_field':
                        self.ignore_none_field = val
                    if name == 'missing_field':
                        if val in ['merge', 'ignore', 'raise_error']:
                            self.missing_field = val
                        else:
                            raise MissingFieldOptionError(
                                f'Option "{val}" is not supported by missing_field. '
                                f'Possible values are ignore, merge, raise_error')

        # Create instance of Meta class and set it to
        # Model class as _meta attribute
        _meta = Meta()
        setattr(cls, '_meta', _meta)

        # Get all fields from model and save them into `_meta.field_list()`
        # and then attach this `_meta` to model class
        for name, field in cls.__dict__.items():
            if isinstance(field, type) and name == 'Meta':
                _meta.set_user_defined_meta(field)
            if isinstance(field, (managers.Manager, fields.Field)):
                field.contribute_to_model(cls, name)

        # Get base Model if they are abstract then add these models field
        # in child classes
        #
        # For example:
        #   class AbstractModel(Model):
        #       name = TextField()
        #
        #       class Meta:
        #           abstract = True
        #
        #   class User(AbstractModel):
        #       age = NumberFiled()
        #
        #   class Student(AbstractModel):
        #       age = NumberFiled()
        #
        # Both Model user and student now contains the name field
        #
        for b in base:
            if not b._meta.abstract:
                from fireo.models import Model
                if b != Model:
                    raise NonAbstractModel(f'Model "{cls.__name__}" can not inherit from non abstract '
                                           f'model "{b.__name__}"')
                continue
            for name, field in b._meta.field_list.items():
                # Ignore the id field
                if isinstance(field, fields.IDField): continue
                field.contribute_to_model(cls, name)

        # Set collection name to model class that is generated from
        # Model class this name can be user defined or auto generated
        # from model class
        #
        # For example:
        #   class UserCollection:
        #       name = TextField()
        #
        # if user not defined collection name then it will auto generate
        # for this class
        #
        # UserCollection will become user_collection
        setattr(cls, "collection_name", cls._meta.collection_name)

        return cls



from fireo.fields.field_attribute import FieldAttribute


class MetaField(type):
    """Get allowed attributes for Fields"""

    def __new__(mcs, name, base, attrs):
        cls = super().__new__(mcs, name, base, attrs)

        allow_attr = cls.allowed_attributes
        for b in base:
            allow_attr += b.allowed_attributes

        setattr(cls, "allowed_attributes", allow_attr)
        return cls


class Field(metaclass=MetaField):
    """Base Field contain common method for all fields

    All fields should be sub class of this `Field` class. Different type of fields handle
    different type of data.

    Attributes
    ----------
    allowed_attributes : list
        Allowed attribute for each fields. This allow to add extra functionality for fields

        Examples
        ---------
        .. code-block:: python
            class User(Model):
                name = TextField(column_name="full_name")

        In firestore this fields will be store as **full_name**

    db_column_name : str
        return the name of the field according to `column_name` attribute if no `column_name` is
        specify then same field name will return

    name:
        name of the field

    model_cls:
        Class of the model

    field_attribute:
        Parse field attributes

    Methods
    -------
    contribute_to_model(mdoel_cls, name):
        Attach the field to model `_meta`

    get_value(val):
        Make validation and return the field db value

    db_value(val):
        How this field is going to save in firestore, Value can be modified here before saving

    field_value(val):
        How this field represent value that is coming from firestore
    """
    allowed_attributes = []

    empty_value_attributes = []

    def __init__(self, *args, **kwargs):
        self.raw_attributes = kwargs
        self.name = None
        self.model_cls = None
        self.field_attribute = FieldAttribute(self, kwargs)

    def contribute_to_model(self, model_cls, name):
        """Attach field to model class

        This method attach field to model class in `_meta` and make this field to `None`

        Parameters
        ----------
        model_cls : Model
            In which model this field will be attached

        name : str
            What is the name of this field when it is attaching with model
        """
        self.name = name
        self.model_cls = model_cls
        setattr(model_cls, name, None)
        model_cls._meta.add_field(self)

    @property
    def db_column_name(self):
        """Get field according to column name

        Return the name of the field according to `column_name` attribute if no `column_name` is
        specify then same field name will return
        """
        return self.raw_attributes.get("column_name") or self.name

    def get_value(self, val, ignore_required=False, ignore_default=False):
        """Get field value after validation

        Make validation and applying attribute function on it.
        it will return the `db_value` how this is going to save in firestore

        Parameters
        ----------
        val : Any
            Field value

        ignore_required : Bool
            Ignore required fields or not mostly ignore when updating the document

        ignore_default : Bool
            Ignore default fields or not mostly ignore when updating the document

        Returns
        -------
            DB value
        """
        val = self.field_attribute.parse(val, ignore_required, ignore_default)
        return self.db_value(val)

    def db_value(self, val):
        """How the value is going to save in firestore

        How this field is going to save in firestore, Value can be modified here before saving
        in firestore.

        Example
        -------
            .. code-block:: python
                class BoolField(Field):
                    def db_value(self, val):
                        if val == True:
                            return 1
                        else:
                            return 0

                class User(Model):
                    is_student = BoolField()

                u = User()
                u.is_student = True

            This will save **1** in firestore

        Returns
        ------
            val:
                Modified value
        """
        if self.model_cls:
            # check if user defined to set the value as lower case
            if self.model_cls._meta.to_lowercase:
                return val.lower() if type(val) is str else val
        return val

    def field_value(self, val):
        """ How this field represent value that is coming from firestore

        Value can be modified after getting value from firestore

        Example
        -------
            .. code-block:: python
                class BoolField(Field):
                    def field_value(self, val):
                        if val == 1:
                            return True
                        else:
                            return False

                class User(Model):
                    is_student = BoolField()

                u = User.collection.get(id="xHuJujlUBkn")
                if u.is_student == True:
                    print("This user is student")

        Returns
        ------
            val:
                Modified value
        """
        return val



from fireo.fields import errors
from fireo.fields.base_field import Field
import re


class TextField(Field):
    """Text field for Models

    Define text for models

    allowed_attributes = ['max_length', 'to_lowercase']



    Examples
    --------
        class User(Model):
            age = TextField()
    """

    allowed_attributes = ['max_length', 'to_lowercase', 'format']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.format_type = None
        self.supported_types = ['title', 'upper', 'lower', 'capitalize']

    def attr_format(self, attr_val, field_val):
        self.format_type = attr_val
        return field_val

    def attr_max_length(self, attr_val, field_val):
        """Method for attribute max_length"""
        return field_val[:attr_val]

    def attr_to_lowercase(self, attr_val, field_val):
        """Method for attribute to_lowercase

            Convert text into lowercase
        """
        if attr_val:
            return field_val.lower() if field_val is not None else None
        return field_val

    def _titlecase(self, s):
        return re.sub(r"[A-Za-z]+('[A-Za-z]+)?",
                      lambda mo: mo.group(0)[0].upper() +
                                 mo.group(0)[1:].lower(),
                      s)

    # override method
    def db_value(self, val):
        if type(val) is str or val is None:
            # check if user defined to set the value as lower case
            if self.model_cls._meta.to_lowercase:
                return val.lower() if val is not None else None
            return val
        raise errors.InvalidFieldType(f'Invalid field type. Field "{self.name}" expected {str}, '
                                      f'got {type(val)}')

    # override method
    def field_value(self, val):

        # check if val is None then there is no need to run these functions
        # just return back the None value
        if val is None:
            return val

        self.field_attribute.parse(val, run_only=['format'])
        if self.format_type:
            if self.format_type in self.supported_types:
                if self.format_type == 'title':
                    return self._titlecase(val)
                if self.format_type == 'upper':
                    return val.upper()
                if self.format_type == 'lower':
                    return val.lower()
                if self.format_type == 'capitalize':
                    return val.capitalize()
            raise errors.AttributeTypeError(
                f'Invalid attribute type. Inside Field "{self.name}", '
                f'"format" type must be one of them "{self.supported_types}".')
        return val



import re

from google.cloud import firestore


def collection_name(model):
    return re.sub('(?!^)([A-Z]+)', r'_\1', model).lower()


def ref_path(key):
    return key.split('/')


def collection_path(key):
    return '/'.join(key.split('/')[:-1])


def get_parent(key):
    return collection_path(key)


def get_parent_doc(key):
    return '/'.join(key.split('/')[:-2])


def get_id(key):
    try:
        return key.split('/')[-1]
    except AttributeError:
        return None


def GeoPoint(latitude: float, longitude: float):
    return firestore.GeoPoint(latitude, longitude)


def get_nested(dict, *args):
    if args and dict:
        element = args[0]
        if element:
            value = dict.get(element)
            return value if len(args) == 1 else get_nested(value, *args[1:])

def generateKeyFromId(model, id):
    return model.collection_name + "/" + id


from fireo.fields import Field, errors


class MapField(Field):
    """Map field for firestore

    Examples
    --------
    .. code-block:: python
        class User(Model):
            marks = MapField()

        u = User()
        u.marks = {'Math': 70, 'English': 80}
    """

    # Override method
    def db_value(self, val):
        if type(val) is dict or val is None:
            # check if user defined to set the value as lower case
            if self.model_cls._meta.to_lowercase:
                return {k: v.lower() if type(v) is str else v for k,v in val.items()}
            return val
        raise errors.InvalidFieldType(f'Invalid field type. Field "{self.name}" expected {dict}, '
                                      f'got {type(val)}')



from fireo.fields import DateTime, NestedModel
from fireo.queries import query_wrapper
from fireo.queries.base_query import BaseQuery
from fireo.utils import utils


class UpdateQuery(BaseQuery):
    """Update document in firestore

    Methods
    ------
    _doc_ref():
        create document ref from firestore

    _parse_field():
        Get and return `db_value` from model `_meta`

    _raw_exec(transaction_or_batch):
        Update document in firestore and return the document

    exec(transaction_or_batch):
        return modified instance of model
    """
    def __init__(self, model_cls, mutable_instance=None, **kwargs):
        super().__init__(model_cls)
        self.query = kwargs
        self.model = mutable_instance
        super().set_collection_path(key=mutable_instance.key)

    def _doc_ref(self):
        """create document ref from firestore"""
        return self.get_ref().document(utils.get_id(self.model.key))

    def _parse_field(self):
        """Get and return `db_value` from model `_meta`

        Examples
        -------
        .. code-block:: python
            class User(Model):
                name = TextField(column_name="full_name")
                age = NumberField()

            User.collection.create(name="Azeem", age=25)

        This method return dict of field names and values
        in this case it will be like this
        `{full_name: "Azeem", age=25}`
        """
        field_dict = {}
        for f in self.model._meta.field_list.values():
            if f.name in self.query:
                # Check if it is nested model
                if isinstance(f, NestedModel):
                    # Get nested model field
                    self._nested_field_list(f, field_dict, f.name)
                else:
                    v = f.get_value(self.query.get(f.name), ignore_required=True, ignore_default=True)
                    if v is not None or type(v) is bool:
                        field_dict[f.db_column_name] = v
                    if v is None and isinstance(f, DateTime):
                        field_dict[f.db_column_name] = v
        return field_dict

    def _nested_field_list(self, f, fl, *name):
        """Get Nested Fields"""
        nested_field_list = {}
        for n_f in f.nested_model._meta.field_list.values():
            if isinstance(n_f, NestedModel):
                n = (*name, n_f.name)
                self._nested_field_list(n_f, nested_field_list, *n)
            else:
                nested_field_list[n_f.db_column_name] = n_f.get_value(
                    utils.get_nested(self.query, *name).get(n_f.name),
                    ignore_required=True
                )
        fl[f.db_column_name] = nested_field_list

    def _raw_exec(self, transaction_or_batch=None):
        """Update document in firestore and return the document"""
        ref = self._doc_ref()
        if transaction_or_batch:
            transaction_or_batch.update(ref, self._parse_field())
            return None

        parse_field = self._parse_field()
        if parse_field:
            ref.update(parse_field)
            return ref.get()

    def exec(self, transaction_or_batch=None):
        """return modified instance of model"""
        if transaction_or_batch:
            return self._raw_exec(transaction_or_batch)
        return query_wrapper.ModelWrapper.from_query_result(self.model, self._raw_exec())



import base64
import json

from fireo.managers.errors import EmptyDocument
from fireo.fields import NestedModel
from fireo.fields.errors import FieldNotFound
from fireo.queries import query_set as queries


class ManagerError(Exception):
    pass


class ManagerDescriptor:
    """Restrict user to get `Manager` from model instance and from abstract model"""

    def __init__(self, manager):
        self.manager = manager

    def __get__(self, instance, owner):
        # reset parent key
        self.manager._parent_key = None
        if instance is not None:
            raise ManagerError(
                f'Manager "{self.manager.name}" can not accessible via {owner.__name__} instance')
        if owner._meta.abstract:
            raise ManagerError(
                f'Manager "{self.manager.name}" is not accessible via {owner.__name__} abstract model')
        return self.manager


class Manager:
    """Manager are used to perform firestore action directly from model class without instance

    Default manager can be accessible via `collection` from model class

    Examples
    -------
    .. code-block:: python
        class User(Model):
            name = TextField()

        user = User.collection.create(name="Azeem")

    Attributes
    ----------
    queryset:
        Read only property, provide operations related to firestore

    _parent_key:
        Parent key if any

    name:
        Name of the manager

    model_cls:
        Model where this manger is contributing

    Methods
    -------
    contribute_to_model(model_cls, name="collection"):
        Attach manager to model class

    create(mutable_instance, kwargs): Model instance
        create new document in firestore collection

    _update(mutable_instance, kwargs): Model instance
        Update existing document in firestore collection

    get(key): Model instance
        Get document from firestore

    get_all(key_list): Model instance
        Get All documents according to key list

    parent(key):
        Parent key if any

    filter(): Model instance
        Get filter document from firestore

    fetch(limit) : generator
        Fetch document from firestore, limit is optional here

    group_fetch(limit) : generator
        Use a collection group query to retrieve documents from a collection group

    transaction():
        Firestore transaction

    batch():
        Firestore batch writes

    limit(count):
        Set limit for query

    offset(num_to_skip)
        Set offset for query

    order(field_name):
        Order document by field_name

    delete(key, child=False)
        Delete document from firestore, key is optional

    delete_all(key_list, batch=None, child=False)
        Delete all documents according to given keys

    cursor(c):
        Start query from specific point

    start_after(key, **kwargs):
        Start document after this key or after that matching fields

    start_at(key, **kwargs):
        Start document at this key or at that matching fields

    end_before(key, **kwargs):
        End document after this key or before that matching fields

    end_at(key, **kwargs):
        End document at this key or at that matching fields
    """

    def __init__(self):
        self.model_cls = None
        self.name = None
        self._parent_key = None

    def contribute_to_model(self, model_cls, name="collection"):
        """Attach manager to model class

        This method attach manager to model class

        Parameters
        ----------
        model_cls : Model
            In which model this manager will be attached

        name : str
            What is the name of this manager when it is attaching with model and
            later can be accessible with this name
        """
        self.name = name
        self.model_cls = model_cls
        setattr(model_cls, name, ManagerDescriptor(self))

    @property
    def queryset(self):
        """provide operations related to firestore"""
        return queries.QuerySet(self.model_cls)

    def create(self, mutable_instance=None, transaction=None, batch=None, merge=None, **kwargs,):
        """create new document in firestore collection

        Parameters
        ---------
        mutable_instance: Model instance
            Make changes in existing model instance After performing firestore action modified this instance
            adding things init like id, key etc

        transaction:
            Firestore transaction

        batch:
            Firestore batch
        """
        _EMPTY_DOC_EXCEPTION = "Empty document can not be save, Add at least one field value"
        # Check if it empty document then don't save it
        if not kwargs:
            raise EmptyDocument(_EMPTY_DOC_EXCEPTION)

        # Check if of the field value is not None
        is_none_dict = True
        for k, v in kwargs.items():
            try:
                f = self.model_cls._meta.get_field(k)
                default_value = f.field_attribute.default
                if v is not None or default_value is not None:
                    is_none_dict = False
                    break
            except FieldNotFound:
                if v is not None or default_value is not None:
                    is_none_dict = False
                    break

        if is_none_dict:
            raise EmptyDocument(_EMPTY_DOC_EXCEPTION)

        field_list = {}
        # if mutable instance is none this mean user is creating document directly from manager
        # For example User.collection.create(name="Azeem") in this case mutable instance will be None
        # If document is creating by directly using manager then check if there is any NestedModel
        # If there is any nested model then get get value from nested model
        if mutable_instance is None:
            for k, v in kwargs.items():
                try:
                    # if this is an id field then save it in field list and pass it
                    f = self.model_cls._meta.get_field(k)
                except FieldNotFound:
                    field_list[k] = v
                    continue
                if isinstance(f, NestedModel):
                    model_instance = v
                    if f.valid_model(model_instance):
                        field_list[f.name] = model_instance._get_fields()
                else:
                    field_list[k] = v
            # Create instance for nested model
            for f in self.model_cls._meta.field_list.values():
                if isinstance(f, NestedModel):
                    field_list[f.name] = f.nested_model()._get_fields()
        else:
            field_list = kwargs

        # If this model has custom IDField then
        # Check if field list length is one(1) and field is IDField
        # if this one field is IDField then this is also Empty Document
        # which can not save
        if self.model_cls._meta.id is not None:
            if len(field_list) == 1:
                # getting first key name from dict
                first_key_name = next(iter(field_list))
                id_name, _ = self.model_cls._meta.id

                # Check first key is id
                if first_key_name == id_name:
                    raise EmptyDocument(_EMPTY_DOC_EXCEPTION)

        return self.queryset.create(mutable_instance, transaction, batch, merge, **field_list)

    def _update(self, mutable_instance=None, transaction=None, batch=None, **kwargs):
        """Update existing document in firestore collection

        Parameters
        ---------
        mutable_instance: Model instance
            Make changes in existing model instance After performing firestore action modified this instance
            adding things init like id, key etc

        transaction:
            Firestore transaction

        batch:
            Firestore batch
        """
        return self.queryset.update(mutable_instance, transaction, batch, **kwargs)

    def get(self, key, transaction=None):
        """Get document from firestore"""
        return self.queryset.get(key, transaction)

    def get_all(self, key_list):
        """Get All documents according to key list"""
        for key in key_list:
            yield self.queryset.get(key)

    def parent(self, key):
        """Parent collection"""
        self._parent_key = key
        return self

    def filter(self, *args, **kwargs):
        """Get filter document from firestore"""
        return self.queryset.filter(self._parent_key, *args, **kwargs)

    def fetch(self, limit=None):
        """Fetch document from collection"""
        return self.queryset.filter(self._parent_key).fetch(limit)

    def group_fetch(self, limit=None):
        """A collection group consists of all collections with the same ID.
        By default, queries retrieve results from a single collection in your database.
        Use a collection group query to retrieve documents from a collection group
        instead of from a single collection."""
        return self.queryset.filter(self._parent_key).group_fetch(limit)

    def transaction(self, t):
        """Firestore transaction"""
        return self.queryset.filter(self._parent_key).transaction(t)

    def batch(self, b):
        """Firestore batch"""
        return self.queryset.filter(self._parent_key).batch(b)

    def limit(self, count):
        """Limit the document"""
        return self.queryset.filter(self._parent_key).limit(count)

    def offset(self, num_to_skip):
        """Set offset for query"""
        return self.queryset.filter(self._parent_key).offset(num_to_skip)

    def order(self, field_name):
        """Order the document by field name"""
        return self.queryset.filter(self._parent_key).order(field_name)

    def delete(self, key=None, transaction=None, batch=None, child=False):
        """Delete document from firestore

        if child is True then delete child collection and documents also
        """
        if key:
            self.queryset.delete(key, transaction, batch, child=child)
        else:
            self.queryset.filter(self._parent_key).delete(child=child)

    def delete_all(self, key_list, batch=None, child=False):
        """Delete all documents according to given keys"""
        for key in key_list:
            self.queryset.delete(key, batch=batch, child=child)

    def cursor(self, cursor):
        """Start query from specific point

        Cursor define where to start the query
        """
        parent = self._parent_key
        cursor_dict = json.loads(base64.b64decode(cursor))
        if 'parent' in cursor_dict:
            parent = cursor_dict['parent']
        query = self.queryset.filter(parent)
        if 'filters' in cursor_dict:
            for filter in cursor_dict['filters']:
                query.filter(*filter)
        if 'order' in cursor_dict:
            query.order(cursor_dict['order'])
        if 'limit' in cursor_dict:
            query.limit(cursor_dict['limit'])

        # check if last doc key is available or not
        if 'last_doc_key' in cursor_dict:
            query.start_after(key=cursor_dict['last_doc_key'])
        else:
            query.offset(cursor_dict['offset'])

        return query

    def start_after(self, key=None, **kwargs):
        """Start document after this key or after that matching fields"""
        return self.queryset.filter(self._parent_key).start_after(key, **kwargs)

    def start_at(self, key=None, **kwargs):
        """Start document at this key or at that matching fields"""
        return self.queryset.filter(self._parent_key).start_at(key, **kwargs)

    def end_before(self, key=None, **kwargs):
        """End document after this key or after that matching fields"""
        return self.queryset.filter(self._parent_key).end_before(key, **kwargs)

    def end_at(self, key=None, **kwargs):
        """End document at this key or at that matching fields"""
        return self.queryset.filter(self._parent_key).end_at(key, **kwargs)



from fireo.fields.errors import *


class FieldAttribute:
    """Parse the field attributes

    Check validation for fields and perform action according to field attributes

    Attributes
    ----------
    allowed_attributes : list
        Allowed attribute for each fields. This allow to add extra functionality for fields

        Examples
        ---------
        .. code-block:: python
            class User(Model):
                name = TextField(column_name="full_name")

        In firestore this fields will be store as **full_name**

    default:
        if no value is define then default value is set for field

    required:
        Required field if no value or default set raise an Error

    validator:
        Custom validation for field specify by user

    Methods
    -------
    validate(value):
        validate the value and perform action according to attribute

    field_attr(attribute):
        Get value of field attribute

    call_attr_method(attr, value):
        Call method from field for custom attribute

    Raises
    ------
    AttributeError:
        if given attributes is not allowed in specific field

    RequiredField:
        If filed is required and no default and no value found

    FieldValidationFailed:
        if field not passed custom user defined validation

    AttributeMethodNotDefined:
        if any custom field not define the method for `allowed_attributes`
    """
    allowed_attributes = ['default', 'required', 'column_name', 'validator']

    def __init__(self, field, attributes):
        self.field = field
        self.attributes = attributes or {}

    # validate each field and it's attributes
    def parse(self, value, ignore_required=False, ignore_default=False, run_only=None):
        """validate the value and perform action according to attribute"""
        for attr in self.attributes:
            if attr not in self.field.allowed_attributes + self.allowed_attributes:
                raise UnSupportedAttribute(f'"{self.field.__class__.__name__}" not support attribute {attr}')

            if run_only is not None:
                for attr in run_only:
                    if self.field_attr(attr) is not None:
                        value = self.call_attr_method(attr, value)

                # return the value back
                return value

            # check default value if set for field
            if self.default is not None and value is None and not ignore_default:
                value = self.default

            # check this field is required or not
            if self.required and value is None and not ignore_required:
                raise RequiredField(f'"{self.field.__class__.__name__}" is required for model {self.field.model_cls} '
                                    f'but received no default and no value.')

            # check if there any custom validation provided by user
            if self.validator is not None:
                if callable(self.validator):
                    # get response back from user defined method
                    validation_passed = self.validator(value)
                    # check type of response
                    if isinstance(validation_passed, bool):
                        if not validation_passed:
                            raise FieldValidationFailed(f'"{self.field.__class__.__name__}" failed validation'
                                                        f' with value {value}')
                    # if response type is tuple then unpack the response
                    # get the user defined error and show to user
                    if isinstance(validation_passed, tuple):
                        valid, error = validation_passed
                        if not valid:
                            raise FieldValidationFailed(f'"{self.field.__class__.__name__}" failed '
                                                        f'with {value}. {error}')
                else:
                    raise ValidatorNotCallable(f'Validator must be a callable, cannot be '
                                               f'{type(self.validator)} {self.validator}')

        # call those attributes method which are defined in this specific field
        # each field can specify any additional attributes
        for attr in self.field.allowed_attributes:
            if self.field_attr(attr) is not None and (value is not None or attr in self.field.empty_value_attributes):
                value = self.call_attr_method(attr, value)

        # return the value back
        return value

    def call_attr_method(self, attr, value):
        """Call method from field for custom attribute

        Any field which is extend from base `Field` and allow it's own attributes
        then attribute method should be defined for each attribute. And attribute
        method should return the value. if no value return then field value will set to None

        And each attribute must be start with the name `attr_` and then the name of attribute itself

            For example:
                for attribute **name** you need to must define method **attr_name()** otherwise it
                will raise `AttributeMethodNotDefined` exception

        Parameters:
        ----------
        attr_val:
            Attribute value that is define in field

        field_val:
            Field value that is define in model

        Return
        ------
        value:
            Modified or just the same value after performing action

        Raises
        ------
        AttributeMethodNotDefined:
            If method is not defined for attribute

        Examples:
        --------
        .. code-block:: python
            class HelloField(Field):
                allowed_attributes = ['if_startfrom']

                def attr_if_startfrom(self, attr_val, field_val)
                    if field_val.startswith(attr_val):
                        return "Hello" + field_val
                    return field_value


            class User(Model):
                name = HelloField(if_startfrom="A")

            u = User()
            u.name = "Azeem"

            # in this case attr_if_startfrom() method run and set the field value
            # as **Hello Azeem**

            u.save()
            print(u.name)  # Hello Azeem

        This is a way how you can create your own custom fields
        """

        try:
            # call attribute method from field
            return getattr(self.field, "attr_"+attr)(self.field_attr(attr), value)
        except AttributeError as e:
            raise AttributeMethodNotDefined(f'Method is not defined for attribute "{attr}" '
                                            f'in field "{self.field.__class__.__name__}"') from e

    def field_attr(self, attr):
        """Get value of field attribute"""
        return self.attributes.get(attr)

    @property
    def default(self):
        """if no value is define then default value is set for field"""
        return self.attributes.get('default')

    @property
    def required(self):
        """Required field if no value or default set raise an Error"""
        return self.attributes.get("required")

    @property
    def validator(self):
        """Custom validation for field specify by user"""
        return self.attributes.get("validator")



from fireo.fields import Field, errors


class BooleanField(Field):
    """Boolean field for firestore

    Examples
    -------
        class User(Model):
            is_student = BooleanField()

        u = User()
        u.is_student = True
    """

    # Override method
    def db_value(self, val):
        if type(val) is bool or val is None:
            return val
        raise errors.InvalidFieldType(f'Invalid field type. Field "{self.name}" expected {bool}, '
                                      f'got {type(val)}')



from fireo import fields
from fireo.managers.managers import Manager
from fireo.models.errors import AbstractNotInstantiate
from fireo.models.model_meta import ModelMeta
from fireo.queries.errors import InvalidKey
from fireo.utils import utils


class Model(metaclass=ModelMeta):
    """Provide easy way to handle firestore features

    Model is used to handle firestore operation easily and provide additional features for best
    user experience.

    Example
    -------
    .. code-block:: python

        class User(Model):
            username = TextField(required=True)
            full_name = TextField()
            age = NumberField()

        user = User()
        user.username = "Axeem"
        user.full_name = "Azeem Haider"
        user.age = 25
        user.save()

        # or you can also pass args into constructor
        user = User(username="Axeem", full_name="Azeem", age=25)
        user.save()

        # or you can use it via managers
        user = User.collection.create(username="Axeem", full_name="Azeem", age=25)

    Attributes
    ----------
    _meta : Meta
        Hold all model information like model fields, id, manager etc

    id : str
        Model id if user specify any otherwise it will create automatically from firestore
        and attached with this model

    key : str
        Model key which contain the model collection name and model id and parent if provided, Id can be user defined
        or generated from firestore

    _update_doc: str
        Update doc hold the key which is used to update the document

    parent: str
        Parent key if user specify

    collection_name : str
        Model name which is saved in firestore if user not specify any then Model class will convert
        automatically in collection name

        For example: UserProfile will be user_profile

    collection : Manager
        Class level attribute through this you can access manager which can be used to save, retrieve or
        update the model in firestore

        Example:
        -------
        .. code-block:: python
            class User(Model):
                name = TextField()

            user = User.collection.create(name="Azeem")

    Methods
    --------
    _get_fields() : dict
        Private method that return values of all attached fields.

    save() : Model instance
        Save the model in firestore collection

    update(doc_key, transaction) : Model instance
        Update the existing document, doc_key optional can be set explicitly

    _set_key(doc_id):
        Set the model key

    Raises
    ------
    AbstractNotInstantiate:
        Abstract model can not instantiate
    """
    # Id of the model specify by user or auto generated by firestore
    # It can be None if user changed the name of id field it is possible
    # to call it from different name e.g user_id
    id = None

    # Private for internal user but there is key property which hold the
    # current document id and collection name and parent if any
    _key = None

    # For sub collection there must be a parent
    parent = ""

    # Hold all the information about the model fields
    _meta = None

    # This is for manager
    collection: Manager = None

    # Collection name for this model
    collection_name = None

    # Track which fields are changed or not
    # it is useful when updating document
    _field_list = []
    _field_changed = []

    # check instance is modified or not
    # When you get the document from firestore or
    # save the document then the model instance changed
    # This also give the help to track update fields
    _instance_modified = False

    # Update doc hold the key which is used to update the document
    _update_doc = None

    def __init__(self, *args, **kwargs):
        # check this is not abstract model otherwise stop creating instance of this model
        if self._meta.abstract:
            raise AbstractNotInstantiate(
                f'Can not instantiate abstract model "{self.__class__.__name__}"')

        # Allow users to set fields values direct from the constructor method
        for k, v in kwargs.items():
            setattr(self, k, v)

        # Create instance for nested model
        # for direct assignment to nested model
        for f in self._meta.field_list.values():
            if isinstance(f, fields.NestedModel):
                if f.name in kwargs:
                    setattr(self, f.name, f.nested_model.from_dict(
                        kwargs[f.name]))
                else:
                    setattr(self, f.name, f.nested_model())

    @classmethod
    def from_dict(cls, model_dict):
        """Instantiate model from dict"""
        if model_dict is None:
            return None
        return cls(**model_dict)

    def to_dict(self):
        """Convert model into dict"""
        model_dict = self._get_fields()
        id = 'id'
        if self._meta.id is not None:
            id, _ = self._meta.id
        model_dict[id] = utils.get_id(self.key)
        model_dict['key'] = self.key
        return model_dict

    # Get all the fields values from meta
    # which are attached with this mode
    # and convert them into corresponding db value
    # return dict {name: value}
    def _get_fields(self):
        """Get Model fields and values

        Retrieve all fields which are attached with Model from `_meta`
        then get corresponding value from model

        Example
        -------
        .. code-block:: python

            class User(Model):
                name = TextField()
                age = NumberField()

            user = User()
            user.name = "Azeem"
            user.age = 25

            # if you call this method `_get_field()` it will return dict{name, val}
            # in this case it will be
            {name: "Azeem", age: 25}

        Returns
        -------
        dict:
            name value dict of model
        """
        field_list = {}
        for f in self._meta.field_list.values():
            if isinstance(f, fields.NestedModel):
                model_instance = getattr(self, f.name)
                if f.valid_model(model_instance):
                    field_list[f.name] = model_instance._get_fields()
            else:
                field_list[f.name] = getattr(self, f.name)
        return field_list

    @property
    def _id(self):
        """Get Model id

        User can specify model id otherwise it will return None and generate later from
        firestore and attached to model

        Example
        --------
        .. code-block:: python
            class User(Mode):
                user_id = IDField()

            u = User()
            u.user_id = "custom_doc_id"

            # If you call this property it will return user defined id in this case
            print(self._id)  # custom_doc_id

        Returns
        -------
        id : str or None
            User defined id or None
        """
        if self._meta.id is None:
            return None
        name, field = self._meta.id
        return field.get_value(getattr(self, name))

    @_id.setter
    def _id(self, doc_id):
        """Set Model id

        Set user defined id to model otherwise auto generate from firestore and attach
        it to with model

        Example:
        --------
            class User(Model):
                user_id = IDField()
                name = TextField()

            u = User()
            u.name = "Azeem"
            u.save()

            # User not specify any id it will auto generate from firestore
            print(u.user_id)  # xJuythTsfLs

        Parameters
        ----------
        doc_id : str
            Id of the model user specified or auto generated from firestore
        """
        id = 'id'
        if self._meta.id is not None:
            id, _ = self._meta.id
        setattr(self, id, doc_id)
        # Doc id can be None when user create Model directly from manager
        # For Example:
        #   User.collection.create(name="Azeem")
        # in this any empty doc id send just for setup things
        if doc_id:
            self._set_key(doc_id)

    @property
    def key(self):
        if self._key:
            return self._key
        try:
            k = '/'.join([self.parent, self.collection_name, self._id])
        except TypeError:
            k = '/'.join([self.parent, self.collection_name, '@temp_doc_id'])
        if k[0] == '/':
            return k[1:]
        else:
            return k

    def _set_key(self, doc_id):
        """Set key for model"""
        p = '/'.join([self.parent, self.collection_name, doc_id])
        if p[0] == '/':
            self._key = p[1:]
        else:
            self._key = p

    def get_firestore_create_time(self):
        """returns create time of document in Firestore

        Returns:
            :class:`google.api_core.datetime_helpers.DatetimeWithNanoseconds`,
            :class:`datetime.datetime` or ``NoneType``:
        """
        return self._meta._firestore_create_time

    def get_firestore_update_time(self):
        """returns update time of document in Firestore

        Returns:
            :class:`google.api_core.datetime_helpers.DatetimeWithNanoseconds`,
            :class:`datetime.datetime` or ``NoneType``:
        """
        return self._meta._firestore_update_time

    def list_subcollections(self):
        """return a list of any subcollections of the doc"""
        if self._meta._referenceDoc is not None:
            return [c.id for c in self._meta._referenceDoc.collections()]

    def save(self, transaction=None, batch=None, merge=None):
        """Save Model in firestore collection

        Model classes can saved in firestore using this method

        Example
        -------
        .. code-block:: python
            class User(Model):
                name = TextField()
                age = NumberField()

            u = User(name="Azeem", age=25)
            u.save()

            # print model id
            print(u.id) #  xJuythTsfLs

        Same thing can be achieved from using managers

        See Also
        --------
        fireo.managers.Manager()

        Returns
        -------
        model instance:
            Modified instance of the model contains id etc
        """
        # pass the model instance if want change in it after save, fetch etc operations
        # otherwise it will return new model instance
        return self.__class__.collection.create(self, transaction, batch, merge, **self._get_fields())

    def upsert(self, transaction=None, batch=None):
        """If the document does not exist, it will be created. 
        If the document does exist it should be merged into the existing document.
        """
        return self.save(transaction=transaction, batch=batch, merge=True)

    def update(self, key=None, transaction=None, batch=None):
        """Update the existing document

        Update document without overriding it. You can update selected fields.

        Examples
        --------
        .. code-block:: python
            class User(Model):
                name = TextField()
                age = NumberField()

            u = User.collection.create(name="Azeem", age=25)
            id = u.id

            # update this
            user = User.collection.get(id)
            user.name = "Arfan"
            user.update()

            print(user.name)  # Arfan
            print(user.age)  # 25

        Parameters
        ----------
        key: str
            Key of document which is going to update this is optional you can also set
            the update_doc explicitly

        transaction:
            Firestore transaction

        batch:
            Firestore batch writes
        """

        # Check doc key is given or not
        if key:
            self._update_doc = key

        # make sure update doc in not None
        if self._update_doc is not None and '@temp_doc_id' not in self._update_doc:
            # set parent doc from this updated document key
            self.parent = utils.get_parent_doc(self._update_doc)
            # Get id from key and set it for model
            setattr(self, '_id', utils.get_id(self._update_doc))
            # Add the temp id field if user is not specified any
            if self._id is None and self.id:
                setattr(self._meta, 'id', ('id', fields.IDField()))
        elif self._update_doc is None and '@temp_doc_id' in self.key:
            raise InvalidKey(
                f'Invalid key to update model "{self.__class__.__name__}" ')

        # Get the updated fields
        updated_fields = {}
        for k, v in self._get_fields().items():
            if k in self._field_changed:
                updated_fields[k] = v
            # Get nested fields if any
            # Nested model store as dict in firestore so check values type is dict
            if type(v) is dict:
                # nested field name and value
                for name, value in v.items():
                    if name in self._field_changed:
                        # create the name with parent field name and child name
                        # For example:
                        #   class User(Model):
                        #       address = TextField()
                        #   class Student(Model):
                        #       age = NumberField()
                        #       user = NestedModel(User)
                        #
                        # Then the field name for nested model will be "user.address"
                        updated_fields[k+"."+name] = value
        # pass the model instance if want change in it after save, fetch etc operations
        # otherwise it will return new model instance
        return self.__class__.collection._update(self, transaction=transaction, batch=batch, **updated_fields)

    def __setattr__(self, key, value):
        """Keep track which filed values are changed"""
        if key in self._field_list or not self._instance_modified:
            self._field_changed.append(key)
        else:
            self._field_list.append(key)
        super(Model, self).__setattr__(key, value)



from fireo.fields import Field, errors
from google.cloud import firestore


class GeoPoint(Field):
    """GeoPoint field for firestore

    Examples
    -------
    .. code-block:: python
        class User(Model):
            location = GeoPoint()

        u = User()
        u.location = fireo.GeoPoint(latitude=123.23, longitude=421.12)
    """

    # Override method
    def db_value(self, val):
        if type(val) is firestore.GeoPoint or val is None:
            return val
        raise errors.InvalidFieldType(f'Invalid field type. Field "{self.name}" expected {firestore.GeoPoint}, '
                                      f'got {type(val)}')


from fireo.queries.base_query import BaseQuery
from fireo.utils import utils
from fireo.database import db


class DeleteQuery(BaseQuery):
    """Delete Document from firestore

    Example
    ------
    .. code-block:: python
        class User(Mode):
            name = TextField()
            age = NumberField()

        u = User(name="Azeem", age=26)
        u.save()

        # Deleting model
        id = User.collection.delete(u.id)
        print(id)

    Methods
    -------
    _delete_collection():
        private method for executing operation in firestore for deleting documents

    exec(transaction_or_batch):
        execute the delete operation
    """

    def __init__(self, model_cls, key=None, query=None, child=False):
        super().__init__(model_cls)
        self.child = child
        self.transaction_or_batch = None
        self.query = query
        self.id = utils.get_id(key)
        if key:
            super().set_collection_path(key=key)

    def _delete_document(self):
        ref = self.get_ref().document(self.id)

        if self.child:
            for c in ref.collections():
                DeleteQuery(self.model_cls, query=c, child=True).exec(self.transaction_or_batch)

        if self.transaction_or_batch:
            self.transaction_or_batch.delete(ref)
        else:
            ref.delete()
        return self.id

    def _delete_collection(self, batch_size=100):
        docs = self.query.limit(batch_size).get()
        deleted = 0

        for doc in docs:
            ref = doc.reference

            if self.child:
                for c in ref.collections():
                    DeleteQuery(self.model_cls, query=c, child=True).exec(self.transaction_or_batch)

            if self.transaction_or_batch:
                self.transaction_or_batch.delete(ref)
            else:
                ref.delete()
            deleted = deleted + 1

        if deleted >= batch_size:
            return self._delete_collection(batch_size)

    def exec(self, transaction_or_batch=None):
        self.transaction_or_batch = transaction_or_batch
        if self.id:
            self._delete_document()
        else:
            self._delete_collection()



from fireo.queries.delete_query import DeleteQuery
from fireo.queries.filter_query import FilterQuery
from fireo.queries.get_query import GetQuery
from fireo.queries.create_query import CreateQuery
from fireo.queries.update_query import UpdateQuery


class QuerySet:
    """Provide operations related to firestore

    Methods
    -------
    create(kwargs):
        Create new document in firestore collection

    update(kwargs):
        Update existing document in firestore collection

    get(key)
        Get document from firestore

    delete(key)
        Delete document in firestore
    """

    def __init__(self, model_cls):
        self.model_cls = model_cls

    def create(self, mutable_instance=None, transaction=None, batch=None, merge=None, **kwargs):
        """Create new document in firestore collection

        Parameters
        ---------
        mutable_instance: Model instance
            Make changes in existing model instance After performing firestore action modified this instance
            adding things init like id, key etc

        transaction:
            Firestore transaction

        batch:
            Firestore batch writes

        **kwargs:
            field name and value

        Returns
        -------
        Model instance:
            modified instance or new instance if no mutable instance provided
        """
        transaction_or_batch = transaction if transaction else batch
        return CreateQuery(self.model_cls, mutable_instance, **kwargs).exec(transaction_or_batch, merge)

    def update(self, mutable_instance=None, transaction=None, batch=None, **kwargs):
        """Update existing document in firestore collection

        Parameters
        ---------
        Parameters
        ---------
        mutable_instance: Model instance
            Make changes in existing model instance After performing firestore action modified this instance
            adding things init like id, key etc

        transaction:
            Firesotre transaction

        batch:
            Firestore batch writes

        **kwargs:
            field name and value

        Returns
        -------
        Model instance:
            updated modified instance
        """
        transaction_or_batch = transaction if transaction else batch
        return UpdateQuery(self.model_cls, mutable_instance, **kwargs).exec(transaction_or_batch)

    def get(self, key, transaction=None):
        """Get document from firestore

        Parameters
        ----------
        key : str
            key of the document

        transaction:
            Firestore transaction

        Returns
        -------
        Model instance:
            wrap query result into model instance
        """
        return GetQuery(self.model_cls, key).exec(transaction)

    def filter(self, parent=None, *args, **kwargs):
        """Filter document from firestore

        Parameters
        ----------
        parent:
            Parent collection if any
        args:
            Where clauses document filter on the base of this

        kwargs:
            keyword args Direct assign for equal filter
        """
        return FilterQuery(self.model_cls, parent, *args, **kwargs)

    def delete(self, key, transaction=None, batch=None, child=False):
        """Delete document from firestore

        Parameters
        ----------
        key : str
            key of the document

        transaction:
            Firestore transaction

        batch:
            Firestore batch writes
        """
        transaction_or_batch = transaction if transaction else batch
        DeleteQuery(self.model_cls, key, child=child).exec(
            transaction_or_batch)



from fireo.fields import NestedModel
from fireo.queries import query_wrapper
from fireo.queries.base_query import BaseQuery
from fireo.utils import utils


class CreateQuery(BaseQuery):
    """Insert model into firestore

    Methods
    ------
    _doc_ref():
        create document ref from firestore

    _parse_field():
        Get and return `db_value` from model `_meta`

    _raw_exec():
        save model into firestore and return the document

    exec():
        return modified or new instance of model
    """

    def __init__(self, model_cls, mutable_instance=None, **kwargs):
        super().__init__(model_cls)
        self.query = kwargs
        # If this is called from manager or mutable model is
        # not provided then this `model` will be a class not instance
        # then create new instance from this model class
        # otherwise set mutable instance to self.model
        if mutable_instance:
            self.model = mutable_instance
            super().set_collection_path(key=mutable_instance.key)
        else:
            self.model = model_cls()

            # Suppose user is defined the id for model
            # let name id **id**
            id_field = 'id'

            # Check user provide any custom name for id
            if model_cls._meta.id is not None:
                id_field, _ = model_cls._meta.id

            # _id setter in model check either user defined
            # any id or not in model
            setattr(self.model, '_id', kwargs.get(id_field))
            # Check if there is any parent
            parent = kwargs.get('parent')
            if parent:
                self.model.parent = parent
                super().set_collection_path(path=parent)
        # Attach key to this model for updating this model
        # Purpose of attaching this key is user can update
        # this model after getting it
        #
        # For example:
        #   u = User.collection.create(name="Azeem", age=25)
        #   u.name = "Updated Name"
        #   u.update()
        self.model._update_doc = self.model.key

        # Reset the field list and changed list
        # This is important to reset so we can
        # find next time which fields are changed
        # when we are going to update it
        self.model._field_list = []
        self.model._field_changed = []

    def _doc_ref(self):
        """create document ref from firestore"""
        return self.get_ref().document(self.model._id)

    def _parse_field(self):
        """Get and return `db_value` from model `_meta`

        Examples
        -------
        .. code-block:: python
            class User(Model):
                name = TextField(column_name="full_name")
                age = NumberField()

            User.collection.create(name="Azeem", age=25)

        This method return dict of field names and values
        in this case it will be like this
        `{full_name: "Azeem", age=25}`
        """
        field_list = {}
        for f in self.model._meta.field_list.values():
            if isinstance(f, NestedModel):
                self._nested_field_list(f, field_list, f.name)
            else:
                fv = f.get_value(self.query.get(f.name))
                if self.model._meta.ignore_none_field:
                    if fv is not None:
                        field_list[f.db_column_name] = fv
                else:
                    field_list[f.db_column_name] = fv
        return field_list

    def _nested_field_list(self, f, fl, *name):
        """Get Nested Fields"""
        required_nested_model = f.raw_attributes.get('required')
        if required_nested_model is None or required_nested_model is False:
            ignore_required = True
        else:
            ignore_required = False
        nested_field_list = {}
        for n_f in f.nested_model._meta.field_list.values():
            if isinstance(n_f, NestedModel):
                n = (*name, n_f.name)
                self._nested_field_list(n_f, nested_field_list, *n)
            else:
                nv = None
                if utils.get_nested(self.query, *name) is not None:
                    nv = utils.get_nested(self.query, *name).get(n_f.name)

                nfv = n_f.get_value(nv, ignore_required)
                if f.nested_model._meta.ignore_none_field:
                    if nfv is not None:
                        nested_field_list[n_f.db_column_name] = nfv
                else:
                    nested_field_list[n_f.db_column_name] = nfv

        fl[f.db_column_name] = nested_field_list

    def _raw_exec(self, transaction_or_batch=None, merge=None):
        """save model into firestore and return the document"""
        ref = self._doc_ref()
        if transaction_or_batch:
            if merge:
                transaction_or_batch.set(
                    ref, self._parse_field(), merge=merge)
            else:
                transaction_or_batch.set(ref, self._parse_field())
            return None

        if merge:
            ref.set(self._parse_field(), merge=merge)
        else:
            ref.set(self._parse_field())
        return ref.get()

    def exec(self, transaction_or_batch=None, merge=None):
        """return modified or new instance of model"""
        if transaction_or_batch:
            return self._raw_exec(transaction_or_batch, merge)
        return query_wrapper.ModelWrapper.from_query_result(self.model, self._raw_exec(merge=merge))



from fireo.database import db
from fireo.fields import NestedModel, DateTime, ReferenceField
from fireo.fields.errors import FieldNotFound
from fireo.queries import query_wrapper
from fireo.queries.base_query import BaseQuery
from fireo.queries.delete_query import DeleteQuery
from fireo.queries.query_iterator import QueryIterator
from fireo.utils import utils
from google.cloud import firestore
from datetime import datetime

# datetime.fromisoformat() is not available before python 3.7
# this package allow to use fromisoformat() method
from backports.datetime_fromisoformat import MonkeyPatch
MonkeyPatch.patch_fromisoformat()


class FilterQuery(BaseQuery):
    """Filter documents on the bases of where clauses

    Filter document by filed name, You can also sort and limit the documents.

    Attributes
    ----------
    model: Model
        Model instance

    select_query: list
        Where filter

    n_limit: int
        Number of documents

    oderer_by: list
        Order the documents

    query_transaction:
        Firestore transaction

    query_batch:
        Firestore batch writes

    Methods
    -------
    transaction(transaction):
        Set Firestore transaction

    batch(batch):
        Set Firestore transaction

    parse_where():
        parse where filter

    query()
        Make a query that perform operation in firestore

    _fields_by_column_name():
        Change the field name according to their db column name

    _firestore_doc():
        Get document from firestore based on key

    start_after(model):
        Start fetching document after the specific model(document)

    start_at(model):
        Start fetching document at the specific model(document)

    end_before(model):
        End fetching document before the specific model(document)

    end_at(model):
        End fetching document at the specific model(document)

    filter(args):
        Apply filter for querying document

    limit(count):
        Apply limit on number of documents

    offset(num_to_skip):
        Define query offset

    order(field_name):
        Order document by field name

    fetch(limit):
        Fetch document from firestore, limit is optional here

    get():
        Get the first matching document from firestore

    delete():
        Delete the filter documents

    _update_doc_key(model):
        Attach key to model for later updating the model
    """

    def __init__(self, model_cls, parent=None, *args, **kwargs):
        super().__init__(model_cls)
        self.model = model_cls()
        self.select_query = self._where_filter(*args, **kwargs)
        self.n_limit = None
        self._offset = None
        self.order_by = []
        self.parent = parent
        self.cursor_dict = {}
        self._start_after = None
        self._start_at = None
        self._end_before = None
        self._end_at = None
        self.query_transaction = None
        self.query_batch = None
        if parent:
            super().set_collection_path(path=parent)
            # Add parent in cursor
            self.cursor_dict['parent'] = parent

    def _where_filter(self, *args, **kwargs):
        """For Direct assign of filter when equality operator"""
        if args:
            return [args]
        elif kwargs:
            filter = []
            for k, v in kwargs.items():
                filter.append((k, '==', v))
            return filter
        else:
            return []

    def transaction(self, t):
        self.query_transaction = t
        return self

    def batch(self, b):
        self.query_batch = b
        return self

    def parse_where(self):
        """Parse where filter

        User can change the field column name in firestore So, filter where clauses
        on the base of db column name
        """
        filters = []
        for w in self.select_query:
            name, op, val = w

            # save the filter in cursor for next fetch
            #
            # ISSUE # 77
            # if filter value type is datetime then it need to first
            # convert into string then JSON serialize
            cf = w
            if type(val) is datetime:
                    cf = (name, op, val.isoformat())

            if 'filters' in self.cursor_dict:
                self.cursor_dict['filters'].append(cf)
            else:
                self.cursor_dict['filters'] = [cf]

            try:
                # ISSUE # 77
                # if field is datetime and type is str (which is usually come from cursor)
                # then convert this string into datetime format
                if isinstance(self.model._meta.get_field(name), DateTime) and type(val) is str:
                    val = datetime.fromisoformat(val)

                # ISSUE # 78
                # check if field is ReferenceField then to query this field we have to
                # convert this value into document reference then filter it
                if isinstance(self.model._meta.get_field(name), ReferenceField):
                    val = db.conn.document(val)
            except FieldNotFound:
                # Filter with nested model not able to find the field (e.g user.name)
                # it require to loop the nested model first to find the field
                # so just ignore it 
                pass

            # check if user defined to set the value as lower case
            if self.model._meta.to_lowercase and type(val) is str:
                val = val.lower()

            # Check it is nested model field
            if '.' in name:
                # m, f = name.split('.')
                # model_field = self.model._meta.get_field(m)
                # model_name = model_field.db_column_name
                # nested_model = model_field.nested_model
                # field_name = nested_model._meta.get_field(f).db_column_name
                # f_name = model_name + '.' + field_name
                model_names = []
                nested_model = None
                *models, field = name.split('.')
                for m in models:
                    try:
                        model_field = self.model._meta.get_field(m)
                    except FieldNotFound:
                        model_field = nested_model._meta.get_field(m)

                    if isinstance(model_field, NestedModel):
                        nested_model = model_field.nested_model
                    name = model_field.db_column_name

                    model_names.append(name)
                model_name = '.'.join(model_names)
                if nested_model is not None:
                    field_name = nested_model._meta.get_field(
                        field).db_column_name
                else:
                    field_name = field
                f_name = model_name + '.' + field_name
            else:
                f_name = self.model._meta.get_field(name).db_column_name
            filters.append((f_name, op, val))
        return filters

    def query(self):
        """Make a query for firestore"""
        ref = self.get_ref()
        # parse where filter
        for f in self.parse_where():
            ref = ref.where(*f)
        # Apply limit
        if self.n_limit:
            ref = ref.limit(self.n_limit)
        # Apply offset
        if self._offset:
            ref = ref.offset(self._offset)
        # order the documents
        for o in self.order_by:
            name, direction = o
            if direction == 'Desc':
                ref = ref.order_by(name, direction=firestore.Query.DESCENDING)
            else:
                ref = ref.order_by(name)

        if self._start_after:
            ref = ref.start_after(self._start_after)
        if self._start_at:
            ref = ref.start_at(self._start_at)
        if self._end_before:
            ref = ref.end_before(self._end_before)
        if self._end_at:
            ref = ref.end_at(self._end_at)

        return ref

    def _fields_by_column_name(self, **kwargs):
        """Change the field name according to their db column name"""
        return {
            self.model._meta.get_field(k).db_column_name: v
            for k,v in kwargs.items()
        }

    def _firestore_doc(self, key):
        """Get document from firestore based on key"""
        return db.conn.collection(utils.get_parent(key)).document(utils.get_id(key)).get()

    def start_after(self, key=None, **kwargs):
        """Start document after this"""
        if key:
            self._start_after = self._firestore_doc(key)
        else:
            self._start_after = self._fields_by_column_name(**kwargs)
        return self

    def start_at(self, key=None, **kwargs):
        """Start document at this point"""
        if key:
            self._start_at = self._firestore_doc(key)
        else:
            self._start_at = self._fields_by_column_name(**kwargs)
        return self

    def end_before(self, key=None, **kwargs):
        """End document before this point"""
        if key:
            self._end_before = self._firestore_doc(key)
        else:
            self._end_before = self._fields_by_column_name(**kwargs)
        return self

    def end_at(self, key=None, **kwargs):
        """End document at this point"""
        if key:
            self._end_at = self._firestore_doc(key)
        else:
            self._end_at = self._fields_by_column_name(**kwargs)
        return self

    def filter(self, *args, **kwargs):
        """Apply filter for querying document

        Apply where filter as many as you want

        Parameters
        ---------
        args: Tuple
            Contain three things 1- field name, 2-operation, 3-value

        kwargs:
            keyword args Direct assign for equal filter

        Returns
        -------
        self:
            Return self object
        """
        if args:
            self.select_query.append(args)
        elif kwargs:
            for k, v in kwargs.items():
                self.select_query.append((k, '==', v))
        return self

    def limit(self, count):
        """Apply limit for query"""
        # save the Limit in cursor for next fetch
        self.cursor_dict['limit'] = count

        if count:
            self.n_limit = count
        return self

    def offset(self, num_to_skip):
        """Offset for query"""
        self._offset = num_to_skip
        return self

    def order(self, field_name):
        """Order document by field name

        By default, a query retrieves all documents that satisfy the query in ascending order by document ID.
        You can specify the sort order for your data using `order()`, and you can limit the number of documents
        retrieved using `limit()`

        Put a dash(-) in front of field name if you want to sort it in descending order. You can also combine
        filter with order

        Parameters
        ----------
        field_name : str
            Name of the field on which base order is applied

        Returns
        -------
            Self object
        """
        # Save order in cursor dict for next fetch
        if 'order' in self.cursor_dict:
            self.cursor_dict['order'] = self.cursor_dict['order'] + ',' + field_name
        else:
            self.cursor_dict['order'] = field_name

        order_direction = 'Asc'
        name = field_name

        # If this is in Desc order
        if field_name[0] == '-':
            order_direction = 'Desc'
            name = field_name[1:]  # Get the field name after dash(-) e.g -age name will be age
        f_name = self.model._meta.get_field(name).db_column_name
        self.order_by.append((f_name, order_direction))
        return self

    def fetch(self, limit=None):
        """Fetch the result from firestore

        Parameters
        ---------
        limit : optional
            Apply limit to firestore documents, how much documents you want to retrieve
        """
        # save the Limit in cursor for next fetch
        self.cursor_dict['limit'] = limit

        if limit:
            self.n_limit = limit
        return QueryIterator(self)

    def group_fetch(self, limit=None):
        super().set_group_collection(True)
        return self.fetch(limit)

    def get(self):
        """Get the first matching document from firestore

        Get first matching document and convert it into model and return it
        This is same as `fetch(limit=1)` the only difference is `get()` method
        return **model instance** and the `fetch()` method return the **generator**
        """
        self.n_limit = 1
        doc = next(self.query().stream(self.query_transaction), None)
        if doc:
            m = query_wrapper.ModelWrapper.from_query_result(self.model, doc)
            m._update_doc = self._update_doc_key(m)
            return m
        return None

    def delete(self, child=False):
        """Delete the filter documents

        if child is True then delete nested collection and documents also
        """
        transaction_or_batch = self.query_transaction if self.query_transaction else self.query_batch
        q = self.query()
        DeleteQuery(self.model, query=q, child=child).exec(transaction_or_batch)

    def _update_doc_key(self, model):
        """Attach key to model for later updating the model

        Attach key to this model for updating this model
        Purpose of attaching this key is user can update
        this model after getting it

        For example:
          u = User.collection.get(user_key)
          u.name = "Updated Name"
          u.update()

        Parameters
        ----------
        model:
            Current model where update key need to attach

        Returns
        -------
        update_doc_key:
            Doc key for updating document
        """
        if self.parent:
            update_doc_key = self.parent + '/' + model.collection_name + '/' + utils.get_id(model.key)
        else:
            update_doc_key = model.key
        return update_doc_key
